/*
 * texdoc.c for W32
 * (2006/02/11 --ak) Public domain.
 *
 * Usage: texdoc docname [SubstringOfPath]
 * 
 * examples:
 * texdoc powerdot
 *        displays powerdot.pdf if it exists under $TEXMF/doc/
 * texdoc README pdftex
 *        displays README under $TEXMF/doc/pdftex/
 * texdoc README dvipdfm
 *        displays README under $TEXMF/doc/dvipdfm/
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>

/*
 * kpathsea functions
 */

#include <kpathsea/kpathsea.h>
#include <windows.h>

/*
 * DVIVIEW:
 * "dviout " for W32TeX
 * "windvi " for TeXLive
 *
 */

#define DVIVIEW "dviout "
/*
#define DVIVIEW "windvi "
*/

char *
k_find_file (char *fname, char *dirsubstr)
{
  char *p, *q;
  char **list, **lp;

  if(!dirsubstr) {
    p = kpse_find_file (fname, kpse_texdoc_format, 0);
    return p;
  } else {
    p = kpse_var_value("TEXDOCS");
    if (!p) {
      fprintf (stderr, "Please define TEXDOCS in texmf.cnf.\n");
      exit (1);
    }
    (void)_strlwr (dirsubstr);
    q = kpse_brace_expand (p);
    free (p);
    list = kpse_all_path_search (q, fname);
    if (q) free (q);
    if (!list) return NULL;
    lp = list;
    for ( ; *lp; lp++) {
      p = *lp;
      (void)_strlwr (p);
      if (strstr (p, dirsubstr)) {
	q = (char *)malloc (strlen (p) + 2);
	strcpy (q, p);
	free (list);
	return q;
      }
    }
  }
  return NULL;
}

int
is_nt (void)
{
  OSVERSIONINFO info;

  info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

  if (!GetVersionEx (&info)) {
    fprintf (stderr, "Failed to get OSVersionInfo.\n");
    exit (1);
  }
  if (info.dwPlatformId == VER_PLATFORM_WIN32_NT) {
    return 1;
  } else {
    return 0;
  }
}

int
run (char *cmd)
{
  int ret;
  ret = win32_system (cmd);
  return ret;
}

int
main (int argc, char **argv)
{
  char *p, *q;
  char *dirsubstr = NULL;
  char cmd[256];
  char fname[126];
  char fullname[256];
  int  prg;

/*
  prg =
  0 : less
  1 : DVIVIEW
  2 : html
  3 : pdf
  4 : ps
*/

  getlongname (argv[0]);
  kpse_set_progname (argv[0]);

  if (argc != 2 && argc != 3) {
    fprintf (stderr, "Usage: %s docname [SubstringOfPath]\n", argv[0]);
    return 1;
  }

  fullname[0] = '\0';

  if (argc == 3) dirsubstr = argv[2];

  strcpy (fname, argv[1]);

  p = k_find_file (fname, dirsubstr);
  if (p) {
    strcpy (fullname, p);
    q = strrchr (p, '.');
    if (!q) {
      prg = 0;
    } else if (!stricmp (q, ".dvi")) {
      prg = 1;
    } else if (!stricmp (q, ".htm")) {
      prg = 2;
    } else if (!stricmp (q, ".html")) {
      prg = 2;
    } else if (!stricmp (q, ".pdf")) {
      prg = 3;
    } else if (!stricmp (q, ".ps")) {
      prg = 4;
    } else if (!stricmp (q, ".eps")) {
      prg = 4;
    } else {
      prg = 0;
    }
    free (p);
  } else {
    strcat (fname, ".txt");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 0;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
    strcpy (fname, argv[1]);
    strcat (fname, ".dvi");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 1;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
    strcpy (fname, argv[1]);
    strcat (fname, ".htm");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 2;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
    strcpy (fname, argv[1]);
    strcat (fname, ".html");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 2;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
    strcpy (fname, argv[1]);
    strcat (fname, ".pdf");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 3;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
    strcpy (fname, argv[1]);
    strcat (fname, ".ps");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 4;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
    strcpy (fname, argv[1]);
    strcat (fname, ".eps");
    p = k_find_file (fname, dirsubstr);
    if (p) {
      prg = 4;
      strcpy (fullname, p);
      free (p);
      goto done;
    }
  }

 done:
  if (!fullname[0]) {
    fprintf (stderr, "I cannot find a document %s", argv[1]);
    if (argc == 3) {
      fprintf (stderr, " in the subdirectories of %s\n", argv[2]);
    } else {
      fprintf (stderr, "\n");
    }
    return 1;
  }

  if (prg == 0) {
    strcpy (cmd, "less ");
  } else if (prg == 1) {
    strcpy (cmd, DVIVIEW);
  } else if (prg == 4) {
    strcpy (cmd, "gsview32 ");
  } else if (prg == 2 || prg == 3) {
    if (is_nt ())
      strcpy (cmd, "cmd /c start ");
    else {
      char *env = getenv("COMSPEC");
      if(env) {
        strcpy(cmd, env);
	strcat(cmd, " /c start ");
      }
      else
	strcpy (cmd, "command /c start ");
    }
  } else {
    fprintf (stderr, "There is no application to view %s\n", argv[1]);
    return 1;
  }
  strcat (cmd, fullname);
  return run (cmd);
}
