/* mktexpk.c (Web2C-7.5.5 --ak 2006)
 * %%----------------------------------------
 * % Special variables for mktexpk ( W32TeX )
 * % ----------------------------------------
 * % MAKETEXPK_STYLE = dosnames
 * % -----------------------------------------
 * % MAKETEXPK_TOP_DIR = $VARTEXFONTS
 * % -----------------------------------------
 * % MAKETEXPK_MODE = canonex
 * % -----------------------------------------
 * % MAKETEXPK_MODE_300 = cx
 * % MAKETEXPK_MODE_400 = nexthi
 * % MAKETEXPK_MODE_600 = canonex
 * % etc.
 * %%-----------------------------------------
 *
 * Usage: mktexpk [OPTIONS] name,
 *   Create a PK font.
 *
 * --dpi DPI           use resolution DPI.
 * --bdpi BDPI         use base resolution BDPI.
 * --mag MAG           use magnificiation MAG.
 * --mfmode MODE       use MODE as the METAFONT mode.
 * --destdir DESTDIR   write fonts in DESTDIR (absolute path).
 *
 * The following old form is also supported:
 * Usage: mktexmk name dpi bdpi mag [mode]
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <direct.h>
#include <io.h>
#include <malloc.h>
#include <process.h>
#include <direct.h>

#include <kpathsea/kpathsea.h>

#define WINWEB2CVERSION "Web2C-7.5.5"
#define WINKPSEVERSION  " Kpathsea-3.5.5"

/* 
 * functions in getdestdir.c
 */

extern char *getdestdir (int, char **);
extern int is_dir (char *);
extern int make_dir (char *);
extern char *testexe (char *);
extern void get_bindir (char *);
extern void mktexupd (char *);

#define LLBUF 1024
#define LBUF  512
#define SBUF  256
#define TBUF  64

/*
Global variables
*/

extern char tex_bindir[];
char *progname;
int UseAspectRatio = 0;
double AspectRatio = 1.0;

void
version (void)
{
  fprintf (stderr, "%s, (C version 1.4 --ak 2006)\n", progname);
  fprintf (stderr, WINWEB2CVERSION WINKPSEVERSION "\n");
}

void
usage (void)
{
  fprintf (stderr, "Usage: %s [OPTIONS] NAME,\n\
  Create a PK font.\n\n\
--dpi DPI           use resolution DPI.\n\
--bdpi BDPI         use base resolution BDPI.\n\
--mag MAG           use magnificiation MAG.\n\
--mfmode MODE       use MODE as the METAFONT mode.\n\
--destdir DESTDIR   write fonts in DESTDIR.\n", progname);
}

void
help (void)
{
  fprintf (stderr, "Usage: %s [OPTIONS] NAME,\n\
  Create a PK font.\n\n\
--dpi DPI           use resolution DPI.\n\
--bdpi BDPI         use base resolution BDPI.\n\
--mag MAG           use magnificiation MAG.\n\
--mfmode MODE       use MODE as the METAFONT mode.\n\
--destdir DESTDIR   write fonts in DESTDIR.\n\n\
Try to create a PK file for NAME at resolution DPI, with an assumed\n\
device base resolution of BDPI, and a Metafont `mag' of MAG. Use MODE\n\
for the METAFONT mode.  Use DESTDIR for the root of where to install\n\
into. DESTDIR must be an absolute directory name. If DESTDIR is not given,\n\
suitable directory is determined according to the TDS.\n\n\
Old form:\n\
Usage: %s name dpi bdpi mag [mode]\n\
is also supported.", progname, progname);
}


void
tpkerr (char *s)
{
  fprintf (stderr, "%s\n", s);
}


void
relmem (char **v)
{
  int j;
  for (j = 0; j < 4; j++)
    free (v[j]);
  return;
}


int
main (int ac, char **av)
{
  HANDLE hnd;
  WIN32_FIND_DATA ffd;

  static char execfile[SBUF];
  char rbuff[LBUF];
  char buff[LBUF];
  char cmd[LBUF];
  char mfname[TBUF];
  char tfname[TBUF];
  char pkname[TBUF];
  char orgav0[SBUF];
  char *pexe;

  char name[TBUF];
  char dpi[TBUF];
  char ydpi[TBUF];
  char bdpi[TBUF];
  char mag[TBUF];
  char mode[TBUF];
  char destdir[SBUF];

  char *arg[4];

  char currdir[SBUF];
  char kpsedot[SBUF];
  char *tmp;
  int cdrive, tdrive;

  FILE *fr, *fw, *fnul, *tfmfileptr;

  int i, savo, savi;
  int style;
  int issetdest;
  int app;
  int oldform;
  int ps2pkok;
  char *env;
  char *p, *fp, *fpp;

  double Xdpi, Ydpi;

  char texname[TBUF], pfbname[TBUF], slant[TBUF], extend[TBUF], encname[TBUF];

/*
 * style =  0 : MAKETEXPK_STYLE undefined or other than dosnames
 * style =  1 : MAKETEXPK_STYLE = dosnames
 */

/*
 * issetdest = 0 : no destdir
 * issetdest = 1 : destdir
 * issetdest = 2 : current working dir
 */

/*
 * app = 0 : mf
 * app = 1 : ps2pk
 * app = 2 : gsftopk
 * app = 3 : ttf2pk
 * app = 4 : hbf2gf
 */

/*
 * oldform = 0 : newform of the command line
 * oldform = 1 : oldform of the command line
 */


/*
 * TEMP | TMP | TMPDIR (necessary)
 *
 */

  tmp = getenv ("TEMP");
  if (!tmp)
    getenv ("TMP");
  if (!tmp)
    getenv ("TMPDIR");
  if (!tmp) {
    tpkerr ("Please define TEMP | TMP | TMPDIR.");
    return (100);
  }

/*
 * normalize directory separators
 */

  for (fpp = tmp; *fpp; fpp++) {
    if (isknj (*fpp)) {
      fpp++;
      continue;
    }
    if (*fpp == '\\')
      *fpp = '/';
  }

  for (i = 0; i < 4; i++)
    arg[i] = (char *) malloc (SBUF);

  strcpy (orgav0, av[0]);
  getlongname (av[0]);
  progname = av[0];
  kpse_set_progname (progname);

  if (ac < 2) {
    usage ();
    relmem (arg);
    return (100);
  }

  issetdest = 0;
  ps2pkok = 0;

/*
 * oldform or newform ?
 *
 */

  if (av[1][0] == '-')
    oldform = 0;
  else
    oldform = 1;


/*
 * Old form of the command line
 */

  if (oldform == 1) {
    if (ac < 5) {
      usage ();
      relmem (arg);
      return (100);
    }
    strcpy (name, av[1]);
    strcpy (dpi, av[2]);
    strcpy (bdpi, av[3]);
    strcpy (mag, av[4]);
    if (ac > 5)
      strcpy (mode, av[5]);
    else
      mode[0] = '\0';
  } else {
/*
 * New form of the command line
 */
    name[0] = dpi[0] = bdpi[0] = mag[0] = mode[0] = destdir[0] = '\0';
    i = 1;
    while (i < ac) {
      if (av[i][0] != '-') {
        strcpy (name, av[i]);
        break;
      }
      if (!strcmp (av[i], "--dpi") || !strcmp (av[i], "-dpi")) {
        i++;
        if (i >= ac) {
          tpkerr ("Invarid arguments.");
          relmem (arg);
          return (100);
        }
        strcpy (dpi, av[i]);
        i++;
      } else if (!strcmp (av[i], "--bdpi") || !strcmp (av[i], "-bdpi")) {
        i++;
        if (i >= ac) {
          tpkerr ("Invarid arguments.");
          relmem (arg);
          return (100);
        }
        strcpy (bdpi, av[i]);
        i++;
      } else if (!strcmp (av[i], "--mag") || !strcmp (av[i], "-mag")) {
        i++;
        if (i >= ac) {
          tpkerr ("Invarid arguments.");
          relmem (arg);
          return (100);
        }
        strcpy (mag, av[i]);
        i++;
      } else if (!strcmp (av[i], "--mfmode") || !strcmp (av[i], "-mfmode")) {
        i++;
        if (i >= ac) {
          tpkerr ("Invarid arguments.");
          relmem (arg);
          return (100);
        }
        strcpy (mode, av[i]);
        i++;
      } else if (!strcmp (av[i], "--destdir") || !strcmp (av[i], "-destdir")) {
        i++;
        if (i >= ac) {
          tpkerr ("Invarid arguments.");
          relmem (arg);
          return (100);
        }
        strcpy (destdir, av[i]);
        issetdest = 1;
        i++;
      } else if (!strcmp (av[i], "--version") || !strcmp (av[i], "-version")) {
        version ();
        relmem (arg);
        return (0);
      } else if (!strcmp (av[i], "--help") || !strcmp (av[i], "-help")) {
        help ();
        relmem (arg);
        return (0);
      } else {
        tpkerr ("Argument error.");
        relmem (arg);
        return (100);
      }
    }
  }                             /* End of command line analysis */

/*
 * set tex_bindir
 */

  get_bindir (orgav0);

  env = kpse_var_value ("MAKETEXPK_STYLE");
  if (!env || (env && strcmp (env, "dosnames"))) {
    style = 0;
  } else
    style = 1;

  if (env) free (env);

/*
 * Default program is mf
 */

  app = 0;

/*
 * determine mfmode if not given
 */

  if (mode[0] == 0 || mode[0] == '/') {
    strcpy (rbuff, "MAKETEXPK_MODE_");
    if (bdpi[0] == 0) {
      tpkerr ("Base DPI cannot be determined.");
      relmem (arg);
      return (100);
    }
    strcat (rbuff, bdpi);

    if (env = kpse_var_value ("MAKETEXPK_MODE"))
      strcpy (mode, env);
    else if (env = kpse_var_value (rbuff))
      strcpy (mode, env);
    else if (!strcmp (bdpi, "118"))
      strcpy (mode, "pcprevw");
    else if (!strcmp (bdpi, "160"))
      strcpy (mode, "nectzo");
    else if (!strcmp (bdpi, "180"))
      strcpy (mode, "lqlores");
    else if (!strcmp (bdpi, "240"))
      strcpy (mode, "canonlbp");
    else if (!strcmp (bdpi, "300"))
      strcpy (mode, "cx");
    else if (!strcmp (bdpi, "320"))
      strcpy (mode, "neclm");
    else if (!strcmp (bdpi, "360"))
      strcpy (mode, "nechi");
    else if (!strcmp (bdpi, "400"))
      strcpy (mode, "nexthi");
    else if (!strcmp (bdpi, "600"))
      strcpy (mode, "canonex");
    else {
      tpkerr ("Cannot determine the mode.");
      tpkerr ("I will try other possibilities.");
      app = 1;
    }
  }

  if (env) free (env);

  if (name[0] == 0) {
    tpkerr ("Font name is not given.");
    relmem (arg);
    return (100);
  }

  if (p = strrchr (name, '.'))
    *p = '\0';

  strcpy (mfname, name);
  strcat (mfname, ".mf");

  if (app == 0) {
    if (!(p = kpse_var_value ("MFINPUTS"))) {
      tpkerr ("Cannot get value of MFINPUTS.");
      relmem (arg);
      return (100);
    }
    free (p);
    if (!(p = kpse_find_file (mfname, kpse_mf_format, 0))) {
      fprintf (stderr, "Cannot find %s .\n", mfname);
      tpkerr ("I try ps2pk --> gsftopk --> ttf2pk --> hbf2gf.");
      app = 1;
    }
  }

  if (app != 0) {
    strcpy (mode, "modeless");
    strcpy (tfname, name);
    strcat (tfname, ".tfm");
    if (!(p = kpse_var_value ("TFMFONTS"))) {
      tpkerr ("Cannot get value of TFMFONTS.");
      relmem (arg);
      return (100);
    }
    free (p);
    if (!(p = kpse_find_file (tfname, kpse_tfm_format, 0))) {
      fprintf (stderr, "Cannot find %s .\n", tfname);
      relmem (arg);
      return 100;
    }
    tfmfileptr = fopen (p, "rb");
    if (!tfmfileptr) {
      fprintf (stderr, "I cannot open %s.\n", p);
      relmem (arg);
      return 100;
    }
    i = 256 * getc (tfmfileptr);
    i += getc (tfmfileptr);
    fclose (tfmfileptr);
    if ((i == 9) || (i == 11)) {
      fprintf (stderr, "Current font seems to be a Japanese one.\n");
      fprintf (stderr, "I give up to create a PK font.\n");
      relmem (arg);
      return 100;
    }
  }

  if ((p[0] == '.') && (p[1] == '/') && (issetdest != 1))
    issetdest = 2;

  fpp = _getcwd (currdir, SBUF);
  if (!fpp) {
    fprintf (stderr, "Failed to get current working directory.\n");
    relmem (arg);
    return (100);
  }
  for (fpp = currdir; *fpp; fpp++) {
    if (isknj (*fpp)) {
      fpp++;
      continue;
    }
    if (*fpp == '\\')
      *fpp = '/';
  }

  i = strlen (currdir);
  if (currdir[i - 1] == '/')
    currdir[i - 1] = '\0';

  strcpy (kpsedot, "KPSE_DOT=.;");
  strcat (kpsedot, currdir);
  _putenv (kpsedot);

  if (issetdest == 2) {
    strcpy (destdir, currdir);
  }

  if (issetdest == 0) {
    strcpy (arg[0], "Dummy");
    strcpy (arg[1], "pk");
    strcpy (arg[2], p);
    strcpy (arg[3], mode);

    if (!(p = getdestdir (4, arg))) {
      tpkerr ("Cannot get destination directory name.");
      relmem (arg);
      return (100);
    }
    strcpy (rbuff, p);
  } else
    strcpy (rbuff, destdir);

/*
 * Change backslash into slash
 */

  for (p = rbuff; *p; p++) {
    if (isknj (*p)) {
      p++;
      continue;
    }
    if (*p == '\\')
      *p = '/';
  }

  p = rbuff;
  i = strlen (p);
  if (p[i - 1] == '/')
    p[i - 1] = '\0';

  if (issetdest) {
    if (!is_dir (p)) {
      fprintf (stderr, "Destination %s is not found.\n", p);
      relmem (arg);
      return (100);
    }
  } else if (!is_dir (p)) {
    if (make_dir (p)) {
      tpkerr ("Error in make_dir.");
      relmem (arg);
      return (100);
    }
  }

  strcpy (buff, p);
  p = buff;

  i = strlen (p);

  if (p[i - 1] != '/')
    strcat (p, "/");

  if (dpi[0] == 0) {
    tpkerr ("Cannot determine DPI.");
    relmem (arg);
    return (100);
  }

  if (style == 1 && issetdest != 2) {   /* dosnames */
    strcat (p, "dpi");
    strcat (p, dpi);
    if (!is_dir (p)) {
      if (make_dir (p)) {
        tpkerr ("Error in make_dir.");
        relmem (arg);
        return (100);
      }
    }
    strcat (p, "/");
  }

  strcat (p, name);
  strcat (p, ".");

  if (style != 1 || issetdest == 2)
    strcat (p, dpi);            /* long suffix */
  strcat (p, "pk");

/* Now buff and p is the full path name of pk file */
/* check the existence of pk file */

  if (_access (p, 0) == 0) {
    fprintf (stderr, "%s exists.\n", p);
    relmem (arg);
    printf ("%s\n", p);
    return (0);
  }

/*
 * Go to the temporary directory
 */

  cdrive = _getdrive ();
  if (tmp[1] == ':') {
    tdrive = tolower (*tmp) - 'a' + 1;
    _chdrive (tdrive);
  }
  _chdir (tmp);

/*
 * save stdout and stdin
 */
  savo = _dup (fileno (stdout));
  savi = _dup (fileno (stdin));

/*
 * connect stdout to stderr
*/
  _dup2 (fileno (stderr), fileno (stdout));

/*
 * connect stdin to nul
 */
  if (!(fnul = fopen ("nul", "rb"))) {
    fprintf (stderr, "Cannot open nul device to read.\n");
    _chdrive (cdrive);
    _chdir (currdir);
    relmem (arg);
    return (100);
  }
  _dup2 (fileno (fnul), fileno (stdin));

/*
 * pkname is the filename of PK font
 */
  sprintf (pkname, "%s.%spk", name, dpi);

  if (app == 0) {
/*
 * METAFONT command line
 */
    if (mag[0] == 0) {
      tpkerr ("Cannot determine MAG.");
      _chdrive (cdrive);
      _chdir (currdir);
      relmem (arg);
      return (100);
    }
    sprintf (cmd,
       "--progname=mf --base=mf \\mode:=%s; \\mag:=%s; nonstopmode; input %s;",
        mode, mag, name);

    pexe = NULL;
    strcpy (execfile, tex_bindir);
    strcat (execfile, "mf-nowin.exe");
    if (!_access (execfile, 0))
      pexe = execfile;
    else {
      fprintf (stderr,
               "%s is not found.\nI search for the file by PATH.\n", execfile);
      pexe = testexe ("mf-nowin.exe");
    }
    if (pexe) {
      fprintf (stderr, "%s %s\n", pexe, cmd);
      (void) spawnlp (P_WAIT, pexe, pexe, cmd, NULL);
    }

    sprintf (cmd, "%s.%sgf", name, dpi);

/*
 * check the consistency
 */
    if (_access (cmd, 0) != 0) {
      tpkerr ("Failed to make gf font by METAFONT.");
      _chdrive (cdrive);
      _chdir (currdir);
      relmem (arg);
      return (100);
    }

/*
 * Change gf into pk
 */
    pexe = NULL;
    strcpy (execfile, tex_bindir);
    strcat (execfile, "gftopk.exe");
    if (!_access (execfile, 0))
      pexe = execfile;
    else {
      fprintf (stderr,
               "%s is not found.\nI search for the file by PATH.\n",
               execfile);
      pexe = testexe ("gftopk.exe");
    }
    if (pexe) {
      fprintf (stderr, "%s %s %s\n", pexe, cmd, pkname);
      (void) spawnlp (P_WAIT, pexe, pexe, cmd, pkname, NULL);
    }

    if (_access (pkname, 0) != 0) {
      tpkerr ("Failed to make pk from gf.");
      _chdrive (cdrive);
      _chdir (currdir);
      relmem (arg);
      return (100);
    }

/*
 * erase gf file
 */
    remove (cmd);

/*
 * erase log file
 */
    sprintf (cmd, "%s.log", name);
    remove (cmd);

/*
 * erase tfm file
 */
    sprintf (cmd, "%s.tfm", name);
    remove (cmd);
    goto finale;
  }

/*
 * app = 1 : ps2pk --> gsftopk --> ttf2pk --> hbf2gf
 */

  p = kpse_find_file ("pspksupp.map", kpse_fontmap_format, 1);
  if (!p) {
    tpkerr ("File pspksupp.map not found.");
    ps2pkok = 0;
    goto do_ps2pk;
  }

  fr = fopen (p, "r");        /* Read pspksupp.map */
  free (p);

  if (!fr) {
    tpkerr ("Cannot open pspksupp.map to read.");
    ps2pkok = 0;
    goto do_ps2pk;
  }

  while (fgets (rbuff, SBUF, fr)) {
    if (rbuff[0] == '%' || rbuff[0] == '#' || rbuff[0] == '\n')
      continue;
    texname[0] = pfbname[0] = slant[0] = extend[0] = encname[0] = '\0';
    i = sscanf (rbuff, "%s %s %s %s %s", texname, pfbname, slant, extend,
		encname);
    if (i == 2 && !strncmp (texname, "AspectRatio", 11)) {
      if (!sscanf (pfbname, "%lf", &AspectRatio)) {
	tpkerr ("File format of pspksupp.map is wrong.");
	fclose (fr);
	ps2pkok = 0;
	goto do_ps2pk;
      }
      UseAspectRatio = 1;
      continue;
    } else if (i > 0 && !stricmp (texname, name)) {
      p = kpse_var_value ("T1FONTS");
      if (!p) {
	tpkerr ("T1FONTS is not defined.");
	ps2pkok = 0;
	break;
      }
      free (p);
      p = kpse_find_file (pfbname, kpse_type1_format, 0);
      if (!p) {
	fprintf (stderr, "%s is not found.\n", pfbname);
	ps2pkok = 0;
	break;
      }
      free (p);
      ps2pkok = 1;
      break;
    }
  }

  fclose (fr);

 do_ps2pk:

  if (ps2pkok) {
    if (UseAspectRatio) {
      sscanf (dpi, "%lf", &Xdpi);
      Ydpi = Xdpi * AspectRatio;
      sprintf (ydpi, "%d", (int) Ydpi);
    } else
      strcpy (ydpi, dpi);

    if (i == 4) {
      sprintf (cmd, "-X%s -Y%s -R%s -S%s -E%s %s %s",
	       dpi, ydpi, bdpi, slant, extend, pfbname, pkname);
    } else if (i == 5) {
      sprintf (cmd, "-e%s -X%s -Y%s -R%s -S%s -E%s %s %s",
	       encname, dpi, ydpi, bdpi, slant, extend, pfbname, pkname);
    } else {
      tpkerr ("File format of pspksupp.map is wrong.");
      goto do_gsftopk;
    }

    pexe = NULL;
    strcpy (execfile, tex_bindir);
    strcat (execfile, "ps2pk.exe");
    if (!_access (execfile, 0))
      pexe = execfile;
    else {
      fprintf (stderr,
	       "%s is not found.\nI search for the file by PATH.\n",
	       execfile);
      pexe = testexe ("ps2pk.exe");
    }
    if (pexe) {
      fprintf (stderr, "%s %s\n", pexe, cmd);
      (void) spawnlp (P_WAIT, pexe, pexe, cmd, NULL);
    }
    if (_access (pkname, 0) != 0) {
      tpkerr ("ps2pk failed to make pk font.");
      goto do_gsftopk;
    }
    goto finale;
  }

/*
 * ps2pk is impossible to use
 */

 do_gsftopk:

  tpkerr ("ps2pk cannot be used.");
  tpkerr ("I try gsftopk.");
  app = 2;

  pexe = NULL;
  strcpy (execfile, tex_bindir);
  strcat (execfile, "gsftopk.exe");
  if (!_access (execfile, 0))
    pexe = execfile;
  else {
    fprintf (stderr,
	     "%s is not found.\nI search for the file by PATH.\n",
	     execfile);
    pexe = testexe ("gsftopk.exe");
  }
  if (pexe) {
    fprintf (stderr, "%s %s %s\n", pexe, name, dpi);
    (void) spawnlp (P_WAIT, pexe, pexe, name, dpi, NULL);
  }
  if (_access (pkname, 0) != 0) {
    tpkerr ("gsftopk cannot be used.");
    tpkerr ("Next I try ttf2pk.");
    app = 3;
    pexe = NULL;
    strcpy (execfile, tex_bindir);
    strcat (execfile, "ttf2pk.exe");
    if (!_access (execfile, 0))
      pexe = execfile;
    else {
      fprintf (stderr,
	       "%s is not found.\nI search for the file by PATH.\n",
	       execfile);
      pexe = testexe ("ttf2pk.exe");
    }
    if (pexe) {
      fprintf (stderr, "%s -q %s %s\n", pexe, name, dpi);
      (void) spawnlp (P_WAIT, pexe, pexe, "-q", name, dpi, NULL);
    }
    if (_access (pkname, 0) != 0) {
      tpkerr ("ttf2pk failed.");
      tpkerr ("Finally I try hbf2gf.");
      app = 4;
      pexe = NULL;
      strcpy (execfile, tex_bindir);
      strcat (execfile, "hbf2gf.exe");
      if (!_access (execfile, 0))
	pexe = execfile;
      else {
	fprintf (stderr,
		 "%s is not found.\nI search for the file by PATH.\n",
		 execfile);
	pexe = testexe ("hbf2gf.exe");
      }
      if (pexe) {
	fprintf (stderr, "%s -q -p %s %s\n", pexe, name, dpi);
	(void) spawnlp (P_WAIT, pexe, pexe, "-q -p", name, dpi, NULL);
      }
      sprintf (cmd, "%s.%sgf", name, dpi);
      if (_access (cmd, 0) != 0) {
	tpkerr ("All trials failed.");
	_chdrive (cdrive);
	_chdir (currdir);
	relmem (arg);
	return (100);
      }
      pexe = NULL;
      strcpy (execfile, tex_bindir);
      strcat (execfile, "gftopk.exe");
      if (!_access (execfile, 0))
	pexe = execfile;
      else {
	fprintf (stderr,
		 "%s is not found.\nI search for the file by PATH.\n",
		 execfile);
	pexe = testexe ("gftopk.exe");
      }
      if (pexe) {
	fprintf (stderr, "%s %s %s\n", pexe, cmd, pkname);
	(void) spawnlp (P_WAIT, pexe, pexe, cmd, pkname, NULL);
      }
      if (_access (pkname, 0) != 0) {
	tpkerr ("All trials failed.");
	_chdrive (cdrive);
	_chdir (currdir);
	relmem (arg);
	return (100);
      }
      remove (cmd);
    }
  }

 finale:

/*
 * return to original stdout and stdin
 */
  _dup2 (savo, fileno (stdout));
  close (savo);
  _dup2 (savi, fileno (stdin));
  close (savi);

/*
 * close nul device
 */
  fclose (fnul);

/*
 * copy the pk file
 */
  if (!(fr = fopen (pkname, "rb"))) {
    fprintf (stderr, "Cannot open %s to read.\n", pkname);
    _chdrive (cdrive);
    _chdir (currdir);
    relmem (arg);
    return (100);
  }

  if (!(fw = fopen (buff, "wb"))) {
    fprintf (stderr, "Cannot open %s to write.\n", buff);
    _chdrive (cdrive);
    _chdir (currdir);
    relmem (arg);
    return (100);
  }

  while (i = fread (rbuff, 1, LBUF, fr))
    fwrite (rbuff, 1, i, fw);

  fclose (fr);
  fclose (fw);
  remove (pkname);

  relmem (arg);

/*
 * update ls-R if it exists
 */
  mktexupd (buff);

/*
 * tell kpathsea
 */

  printf ("%s\n", buff);
  _chdrive (cdrive);
  _chdir (currdir);

  return (0);
}
