################################################################################
#
# Makefile  : eTeX, web2c win32.mak makefile fragment to build e-TeX
# Author    : Fabrice Popineau <Fabrice.Popineau@supelec.fr>
# Platform  : Win32, Microsoft VC++ 6.0, depends upon fpTeX 0.5 sources
# Time-stamp: <02/11/09 22:44:23 popineau>
#
################################################################################

etexver = 

# The C sources.
etex_c = etex.c
etex_o = $(objdir)\etex.obj

# Generation of the web and ch file.
etex.web: $(objdir)\tie.exe tex.web etexdir/etex.ch etexdir/etex.fix
	.\$(objdir)\tie -m etex.web		\
		           $(srcdir)/tex.web	\
			   $(srcdir)/etexdir/etex.ch \
			   $(srcdir)/etexdir/etex.fix

etex.ch: $(objdir)\tie.exe etex.web etexdir/tex.ch0 tex.ch tex-supp-w32.ch etexdir/tex.ch1 etexdir/tex.ech etexdir/tex.ch2
	.\$(objdir)\tie -c etex.ch etex.web			\
			   $(srcdir)/etexdir/tex.ch0 	\
			   $(srcdir)/tex.ch	 		\
			   $(srcdir)/tex-supp-w32.ch	 		\
			   $(srcdir)/etexdir/tex.ch1 	\
			   $(srcdir)/etexdir/tex.ech	\
			   $(srcdir)/etexdir/tex.ch2

# Tests...
testdir = $(srcdir)\triptrap
etestdir = $(srcdir)\etexdir\etrip
dvitype_args = -output-level=2 -dpi=72.27 -page-start=*.*.*.*.*.*.*.*.*.*

etrip: $(objdir)\pltotf.exe $(objdir)\tftopl.exe $(objdir)\etex.exe $(objdir)\dvitype.exe
	@echo ">>> See $(etestdir)/etrip.diffs for example of acceptable diffs."
	@echo "*** TRIP test for e-TeX in compatibility mode ***."
	.\$(objdir)\pltotf $(testdir)/trip.pl trip.tfm
	.\$(objdir)\tftopl ./trip.tfm trip.pl
	-$(diff) $(testdir)/trip.pl trip.pl
# get same filename in log
	-$(del) /f trip.tex & $(copy) $(testdir)\trip.tex . 
# FIXME: this works only written on 2 lines !!!
	set TEXMFCNFOLD=$(TEXMFCNF)
	set TEXMFCNF=$(etestdir)
	-.\$(objdir)\etex --progname=einitex <$(testdir)\trip1.in >ctripin.fot
	$(copy) trip.log ctripin.log & $(del) trip.log
	-$(diff) $(testdir)/tripin.log ctripin.log
	-.\$(objdir)\etex <$(testdir)\trip2.in >ctrip.fot
	$(copy) trip.log ctrip.log & $(del) trip.log
	-$(diff) $(testdir)/trip.fot ctrip.fot
	-$(diff) $(diffflags) $(testdir)/trip.log ctrip.log
	.\$(objdir)\dvitype $(dvitype_args) trip.dvi >ctrip.typ
	-$(diff) $(diffflags) $(testdir)/trip.typ ctrip.typ
	@echo "*** TRIP test for e-TeX in extended mode ***."
	-.\$(objdir)\etex -progname=einitex <$(etestdir)\etrip1.in >xtripin.fot
	$(copy) trip.log xtripin.log & $(del) trip.log
	-$(diff) ctripin.log xtripin.log
	-.\$(objdir)\etex <$(etestdir)\trip2.in >xtrip.fot
	$(copy) trip.log xtrip.log & $(del) trip.log
	-$(diff) ctrip.fot xtrip.fot
	-$(diff) $(diffflags) ctrip.log xtrip.log
	.\$(objdir)\dvitype $(dvitype_args) trip.dvi >xtrip.typ
	-$(diff) $(diffflags) ctrip.typ xtrip.typ
	@echo "*** e-TeX specific part of e-TRIP test ***."
	.\$(objdir)\pltotf $(etestdir)/etrip.pl etrip.tfm
	.\$(objdir)\tftopl ./etrip.tfm etrip.pl
	-$(diff) $(etestdir)/etrip.pl etrip.pl
# get same filename in log
	-$(del) /f etrip.tex & $(copy) $(etestdir:/=\)\etrip.tex . 
	-.\$(objdir)\etex --progname=einitex <$(etestdir)\etrip2.in >etripin.fot
	$(copy) etrip.log etripin.log & $(del) etrip.log
	-$(diff) $(etestdir)/etripin.log etripin.log
	-.\$(objdir)\etex <$(etestdir)\etrip3.in >etrip.fot
	-$(diff) $(etestdir)/etrip.fot etrip.fot
	-$(diff) $(diffflags) $(etestdir)/etrip.log etrip.log
	set TEXMFCNF=$(TEXMFCNFOLD)

etex-check: etrip etex.efmt
# Test truncation (but don't bother showing the warning msg).
	-.\$(objdir)\etex --output-comment="$(outcom)" \
	  $(srcdir)/tests/hello \
	  && .\$(objdir)\dvitype hello.dvi | grep kb@mail.tug.org >nul
# \openout should show up in \write's.
	-.\$(objdir)\etex $(srcdir)/tests/openout && grep xfoo openout.log
# one.two.tex -> one.two.log
	-.\$(objdir)\etex $(srcdir)/tests/one.two && ls -l one.two.log
# uno.dos -> uno.dos.log
	-.\$(objdir)\etex $(srcdir)/tests/uno.dos && ls -l uno.dos.log
	-.\$(objdir)\etex $(srcdir)/tests/just.texi && ls -l just.log
	-.\$(objdir)\etex $(srcdir)/tests/batch.tex
	-.\$(objdir)\etex --shell $(srcdir)/tests/write18 | grep echo
# tcx files are a bad idea.
#	.\$(objdir)\etex --translate-file=$(srcdir)/share/isol1-t1.tcx \
#	  $(srcdir)/tests/eight && ./dvitype eight.dvi >eigh.typ
	-.\$(objdir)\etex --mltex --progname=einitex $(srcdir)/tests/mltextst
	-.\$(objdir)\etex <nul
#	-PATH=$(kpathsea_dir):$$PATH .\$(objdir)\etex '\nonstopmode\font\foo=nonesuch\end'
	set PATH=$(kpathsea_dir);$(kpathsea_srcdir);$(PATH)
	set MKTEXSCRIPT=$(kpathsea_srcdir)
	set TMPDIR=..
	-.\$(objdir)\etex "\nonstopmode\font\foo=nonesuch\end"

# End of etex.mk.
# 
# Local variables:
# page-delimiter: "^# \f"
# mode: Makefile
# End:
